"""Test parsing CLI XML
"""

import os
import sys

import lxml.etree
import requests
import lxml
import json

parameter_tags = ['integer','float','double','boolean','string','integer-vector','float-vector','double-vector','string-vector',
                  'integer-enumeration','float-enumeration','double-enumeration','string-enumeration','file','directory','image',
                  'geometry','point','pointfile','region','table','transform']

# Additional options (which don't appear on a lot of the plugins we've made)
# constraints (not for use in vector or enumeration)
#   # includes minimum, maximum, step (which can be 0)

# slicer schema includes examples of text output from CLI, (specifying --returnparameterfile file_name)
# but I don't think that goes anywhere on Girder

def main():

    api_url = 'http://ec2-3-230-122-132.compute-1.amazonaws.com:8080/api/v1/'
    plugin_id = '63e6bc1da00b00eade3047c1'
    girder_token = ''
    xml_req = requests.get(
        api_url+f'slicer_cli_web/cli/{plugin_id}/xml?token={girder_token}'
    )

    xml_data = lxml.etree.fromstring(xml_req.content)
    print(f"Number of parameters: {len(xml_data.findall('parameters'))}")
    print(xml_data.find('title').text)
    print(xml_data.find('description').text)
    xml_params = xml_data.findall('parameters')
    
    # Parameters are broken into groups with names (first one is IO)
    for param in xml_data.iterfind('parameters'):
        # .keys() on an element returns the "attrib" names
        #param_attribs = param.keys()
        # .items() returns (name,value) sequence of "attrib"s
        # .get(key,default=None) will return the value of the "attrib" "key" if it's present, otherwise will return "default"
        is_advanced = param.get('advanced',default=False)
        if is_advanced:
            print(f'{param.find("label").text} is an advanced set of parameters!')
        else:
            print(f'{param.find("label").text} is not an advanced set of parameters!')

        print(param.find('description').text)
        print('---------------------------------')
        for sub_el in param:
            if sub_el.tag in parameter_tags:
                print(f'Input param: {sub_el.find("label").text} is of the type: {sub_el.tag}')
                default_value = sub_el.find('default')
                if not default_value is None:
                    print(f'default value: {default_value.text}')
                else:
                    print('No default value')

                if 'enumeration' in sub_el.tag:
                    print('Enumeration options:')
                    for opt_idx,opt in enumerate(sub_el.iterfind('element')):
                        print(f'Option: {opt_idx}: {opt.text}')


if __name__=='__main__':
    main()







































"""Loading Visium annotations from CSV
"""

import os
import sys
sys.path.append('./src/')
import json

from fusion_tools.utils.shapes import load_visium

def main():
    
    # 
    base_dir = 'C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\KPMP_Atlas_V2\\'
    coords_path = base_dir+'V10S14-085_XY01_20-0038\\V10S14-085_XY01_20-0038_spot_info.csv'
    scale_factors = coords_path.replace('_spot_info.csv','_scalefactors.json')

    with open(scale_factors,'r') as f:
        scales = json.load(f)
        f.close()

    visium_coords = load_visium(
        coords_path,
        include_var_names=['DCT1'],
        scale_factor=scales['lowres']
    )
    print(len(visium_coords['features']))
    print(visium_coords['features'][0]['properties'])



if __name__=='__main__':
    main()



"""Base classes for tools
"""
from typing_extensions import Union

class Tool:
    """General class for interactive components that visualize, edit, or perform analyses on data.
    """
    def __init__(self):
        # Property referring to how the layout is updated with a change in the 
        # visualization session
        self.session_update = False

    def prefix_escape(self,input_id: Union[str,dict])->bool:
        """Specifying the default prefix escape for all blueprints. Basic specification is to ignore "anchor" or ids that already have a prefix

        :param input_id: Input component id either for a layout or a callback
        :type input_id: Union[str,dict]
        :return: True indicates that this id should not receive a prefix transform, False indicates that it should receive a prefix
        :rtype: bool
        """

        if type(input_id)==dict:
            if 'anchor' in input_id['type']:
                return True
            try:
                current_comp_id = int(input_id['type'][0])
                has_id = True
            except ValueError:
                has_id = False
            
            if has_id:
                return True

        elif 'anchor' in input_id:
            return True
        return False
    

class MultiTool:
    """General class for a Tool which works on multiple slides at once
    """
    def __init__(self):
        # Property referring to how the layout is updated with a change in the 
        # visualization session
        self.session_update = True

    def prefix_escape(self,input_id: Union[str,dict])->bool:
        """Specifying the default prefix escape for all blueprints. Basic specification is to ignore "anchor" or ids that already have a prefix

        :param input_id: Input component id either for a layout or a callback
        :type input_id: Union[str,dict]
        :return: True indicates that this id should not receive a prefix transform, False indicates that it should receive a prefix
        :rtype: bool
        """

        if type(input_id)==dict:
            if 'anchor' in input_id['type']:
                return True
            try:
                current_comp_id = int(input_id['type'][0])
                has_id = True
            except ValueError:
                has_id = False
            
            if has_id:
                return True

        elif 'anchor' in input_id:
            return True
        return False
    
    
class Handler:
    pass

class DSATool(Tool):
    """A sub-class of Tool specific to DSA components. 
    The only difference is that these components always update relative to the session data.

    :param Tool: General class for components that perform visualization and analysis of data.
    :type Tool: None
    """
    def __init__(self):

        super().__init__()
        self.session_update = True


class MapComponent:
    """General class for components added to SlideMap
        For more information see dash-leaflet: https://www.dash-leaflet.com/

    """
    def __init__(self):
        # Property referring to how the layout is updated with a change in the 
        # visualization session
        self.session_update = True
    
    def prefix_escape(self,input_id: Union[str,dict])->bool:
        """Specifying the default prefix escape for all blueprints. Basic specification is to ignore "anchor" or ids that already have a prefix

        :param input_id: Input component id either for a layout or a callback
        :type input_id: Union[str,dict]
        :return: True indicates that this id should not receive a prefix transform, False indicates that it should receive a prefix
        :rtype: bool
        """

        if type(input_id)==dict:
            if 'anchor' in input_id['type']:
                return True
            try:
                current_comp_id = int(input_id['type'][0])
                has_id = True
            except ValueError:
                has_id = False
            
            if has_id:
                return True

        elif 'anchor' in input_id:
            return True
        return False
    




"""

Using the SegmentationDataset for segmenting cells using cellpose

"""
# This is for running with pipx environment, $ pipx run test_segmentation.py
# /// script
# dependencies = ["cellpose","fusion-tools[interactive]"]
# ///

with_predictions = False

import os
import sys
sys.path.append('./src/')
import numpy as np
import pandas as pd
import geopandas as gpd
import json

from fusion_tools.dataset import SegmentationDataset
#from wsi_annotations_kit import wsi_annotations_kit as wak

#import large_image
from tqdm import tqdm
if with_predictions:
    from cellpose import io, models, utils

import rasterio
import rasterio.features

from fusion_tools.visualization import Visualization
from fusion_tools.components import SlideMap, OverlayOptions

import uuid


def mask_to_shape(mask: np.ndarray, bbox: list)->list:

    mask_features = []
    for geo, val in rasterio.features.shapes(mask, mask = mask>0):
        mask_features.append({
            'type': 'Feature',
            'geometry': {
                'type': geo['type'],
                'coordinates': [[
                    [float(i[0]+bbox[0]),float(i[1]+bbox[1])]
                    for i in geo['coordinates'][0]
                ]]
            },
            'properties': {
                'name': 'Segmented Cell'
            }
        })
    
    return mask_features


def main():

    slides = [
        #'C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\Cropped_15-1 new merged.tiff'
        'C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\XY01_IU-21-015F.svs'
    ]

    seg_dataset = SegmentationDataset(
        slides = slides,
        annotations = None,
        use_parallel=False,
        verbose = True,
        patch_mode = 'all',
        patch_region = 'tissue',
        patch_size = [2*224,2*224]
    )

    print(seg_dataset.slide_data[0]['metadata'])
    print(f'Number of patches = {len(seg_dataset)}')
    
    if with_predictions:
        cell_model = models.CellposeModel(model_type='tissuenet_cp3',gpu=False)

    all_cells_gdf = gpd.GeoDataFrame()
    if with_predictions:
        with tqdm(seg_dataset, total = len(seg_dataset)) as pbar:
            pbar.set_description('Predicting on patches in SegmentationDataset')
            for idx, (patch,_) in enumerate(seg_dataset):
                
                masks, _, _ = cell_model.eval(patch, diameter=10, cellprob_threshold = 0.0, channels=[0,0])


                if max(np.unique(masks))>0:
                    # Converting masks to annotations
                    mask_bbox = seg_dataset.data[idx]['bbox']
                    mask_geos = mask_to_shape(masks,mask_bbox)
                    if all_cells_gdf.empty:
                        all_cells_gdf = gpd.GeoDataFrame.from_features(mask_geos)

                    else:
                        new_cells = gpd.GeoDataFrame.from_features(mask_geos)
                        all_cells_gdf = pd.concat([all_cells_gdf,new_cells],axis=0,ignore_index=True)
                        merged_geoms = all_cells_gdf.union_all().geoms
                        all_cells_gdf = gpd.GeoDataFrame({'geometry': merged_geoms, 'name': ["Segmented Cells"]*len(merged_geoms)})

                pbar.update(1)

        pbar.close()
    
    # Adding the bounding boxes for patches:
    bbox_geos = {
        'type': 'FeatureCollection',
        'features': [
            {
                'type': 'Feature',
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[
                        [i['bbox'][0],i['bbox'][1]],
                        [i['bbox'][2],i['bbox'][1]],
                        [i['bbox'][2],i['bbox'][3]],
                        [i['bbox'][0],i['bbox'][3]],
                        [i['bbox'][0],i['bbox'][1]]
                    ]],
                },
                'properties': {
                    'name': 'Bounding Boxes',
                    '_id': uuid.uuid4().hex[:24]
                }
            }
            for i in seg_dataset.data
        ],
        'properties': {
            'name': 'Bounding Boxes',
            '_id': uuid.uuid4().hex[:24]
        }
    }

    
    if with_predictions:
        all_cells_geo = all_cells_gdf.to_geo_dict(show_bbox=False)
        all_cells_geo['properties'] = {'name': 'Segmented Cells','_id': uuid.uuid4().hex[:24]}
        with open(slides[0].replace('tiff','json'),'w') as f:
            json.dump(all_cells_geo,f)

            f.close()

        local_annotations = [[bbox_geos,all_cells_geo]]
    else:
        local_annotations = [[bbox_geos]]

    local_slides = slides

    vis = Visualization(
        local_slides=local_slides,
        local_annotations=local_annotations,
        components = [
            SlideMap(),
            [
                OverlayOptions()
            ]
        ]
    )
    vis.start()

if __name__=='__main__':
    main()



"""

Saving session information on a linked DSA instance

"""

import json

from typing_extensions import Union

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import callback, ctx, ALL, MATCH, exceptions, no_update, dcc
import dash_bootstrap_components as dbc
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform

from fusion_tools.visualization.vis_utils import get_pattern_matching_value
from fusion_tools import DSATool


class DSASession(DSATool):
    """Handler for saving a session up on a linked DSA instance

    :param DSATool: Sub-class of Tool specific to DSA components. Updates with session data by default.
    :type DSATool: _type_
    """
    def __init__(self,
                 handler):
        
        self.handler = handler

    def __str__(self):
        return 'DSA Session'

    def load(self, component_prefix:int):
        self.component_prefix = component_prefix

        self.title = 'DSA Session'
        self.blueprint = DashBlueprint(
            transforms=[
                PrefixIdTransform(prefix=f'{component_prefix}'),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

    def update_layout(self, session_data:dict, use_prefix:bool):
        
        layout = html.Div([
            html.H4(
                children = [
                    f'Click the button to save your current session on DSA'
                ]
            ),
            html.Hr(),
            html.Div(
                dbc.Stack([
                    dbc.Button(
                        'Save Session',
                        className = 'd-grid col-12 mx-auto',
                        color = 'primary',
                        id = {'type': 'dsa-save-session-button','index': 0},
                        n_clicks = 0
                    ),
                    dcc.Loading(
                        html.Div(
                            id = {'type': 'dsa-session-status-div','index': 0},
                            children = []
                        )
                    )
                ],direction = 'vertical')
            )
        ],style = {'padding': '10px 10px 10px 10px'})

        if use_prefix:
            PrefixIdTransform(prefix = self.component_prefix).transform_layout(layout)

        return layout

    def gen_layout(self, session_data:dict):
        """Creating the layout for this component.

        :param session_data: Dictionary containing current session info
        :type session_data: dict
        """

        self.blueprint.layout = self.update_layout(session_data,use_prefix = False)

    def get_callbacks(self):
        
        self.blueprint.callback(
            [
                Input({'type': 'dsa-save-session-button','index': ALL},'n_clicks')
            ],
            [
                State('anchor-vis-store','data'),
                State('anchor-page-url','href'),
                State('anchor-page-url','pathname')
            ],
            [
                Output({'type': 'dsa-session-status-div','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.save_session)

    def save_session(self, clicked, session_data, page_url, pathname):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)

        if not 'current_user' in session_data:
            session_status_div = html.Div(
                dbc.Alert(
                    'Sign in first to save a session',
                    color = 'danger',
                    dismissable=True
                )
            )
            return [session_status_div]

        saved_session_data = {
            'page_url': page_url,
            'page': pathname,
            'current': session_data['current'],
            'data': session_data['data'],
            'user_session': session_data['current_user']['login']
        }

        uploaded_file_details = self.handler.upload_session(saved_session_data,user_token = session_data['current_user']['token'])
        
        # Find a way to extract the window url or something
        session_link = f'{page_url}/session?id={uploaded_file_details["_id"]}'

        session_status_div = html.Div([
            dbc.Alert('Session Saved Successfully!',color = 'success',dismissable=True),
            dbc.Row([
                dbc.Col(
                    dbc.Label(
                        'Session URL:',
                        style = {'fontSize': 15}
                    ),
                    md = 2
                ),
                dbc.Col([
                    dcc.Input(
                        value = session_link,
                        disabled = True,
                        id = {'type': f'{self.component_prefix}-dsa-session-link','index': 0},
                        style = {
                            'width': '100%',
                            'fontSize': 20
                        }
                    )
                ], md = 9),
                dbc.Col(
                    dcc.Clipboard(
                        target_id = {'type': f'{self.component_prefix}-dsa-session-link','index':0},
                        title = 'copy',
                        style = {
                            'fontSize': 20,
                            'display': 'inline-block'
                        }
                    ),
                    md = 1
                )
            ])
        ])

        return [session_status_div]





from fusion_tools.components.maps import MapComponent, SlideMap, MultiFrameSlideMap, SlideImageOverlay, ChannelMixer, LargeSlideMap, LargeMultiFrameSlideMap
from fusion_tools.components.tools import OverlayOptions, PropertyViewer, PropertyPlotter, HRAViewer, DataExtractor, GlobalPropertyPlotter
from fusion_tools.components.segmentation import FeatureAnnotation, BulkLabels



"""

Testing tile server with local tile source

"""

import os
import sys
sys.path.append('./src/')
from fusion_tools.visualization.components import LocalTileServer
import threading


def main():

    path_to_slide = 'C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\XY01_IU-21-015F.svs'

    tile_server = LocalTileServer(
        local_image_path = [path_to_slide],
        tile_server_port=8080
    )
    print(tile_server)
    #TODO: Should this be included in the tile_server.start method?
    new_thread = threading.Thread(
        target = tile_server.start,
        name = 'local_tile_server',
        daemon=True
    )
    new_thread.start()


if __name__=='__main__':
    main()





"""Testing out different dataset classes
"""

import os
import sys
sys.path.append('./src/')
from time import sleep

import large_image
import numpy as np
from fusion_tools.dataset import SegmentationDataset, ClassificationDataset
from fusion_tools.utils.shapes import load_aperio
import matplotlib.pyplot as plt

def main():
    
    test_slide_path = "C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\XY01_IU-21-015F_001.svs"
    slide_list = [
        test_slide_path
    ]
    annotations_list = [
        load_aperio(test_slide_path.replace('svs','xml'))    
    ]

    # Default parameters
    seg_dataset = SegmentationDataset(
        slides = slide_list,
        annotations = annotations_list,
        patch_size = [512,512],
        patch_mode = 'centered_bbox'
    )

    # Printing key_configs
    print(seg_dataset)
    print(f'len(seg_dataset): {len(seg_dataset)}')

    # Viewing some image/mask combos:
    for i in range(len(seg_dataset)):
        image, mask = seg_dataset[i]

        plt.imshow(image)
        plt.show(block=False)
        plt.pause(0.25)
        plt.close('all')

        # Normalizing masks (channel = class)
        mask = np.sum(mask,axis=-1)
        plt.imshow(255*mask)
        plt.show(block=False)
        plt.pause(0.25)
        plt.close('all')


    class_dataset = ClassificationDataset(
        slides = slide_list,
        annotations=annotations_list,
        label_property = 'name'
    )

    print(class_dataset)
    print(f'len of class_dataset: {len(class_dataset)}')

    for i in range(len(class_dataset)):
        image,label = class_dataset[i]

        print(f'label: {label}')
        plt.imshow(image)
        plt.show(block=False)
        plt.pause(0.5)
        plt.close('all')

if __name__=='__main__':
    main()




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fusion-tools",
    version="3.5.28",
    author="Sam Border",
    author_email="sam.border2256@gmail.com",
    description="Modular visualization and analysis dashboard creation for high-resolution microscopy images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/spborder/fusion-tools",
    python_requires=">=3.8",
    install_requires=[
        "lxml>=4.9.2",
        "geojson>=3.0.1",
        "shapely>=2.0.1",
        "anndata",
        "geopandas>=1.0.1",
        "girder-client",
        "numpy>=1.20.0",
        "large-image[common]",
        "requests",
        "uuid>=1.30",
        "scikit-image",
        "umap-learn>=0.5.6",
        "statsmodels>=0.14.0",
        "typing-extensions>=4.8.0",
        "girder-job-sequence>=0.2.7"
    ],
    extras_require = {
        'interactive': [
            "dash-leaflet[all]>=1.0.15",
            "dash>=2.18.1,<3.0.0",
            "dash-extensions>=1.0.18",
            "dash-uploader==0.7.0-a1",
            "dash_mantine_components>=0.14.4",
            "dash-bootstrap-components>=1.6.0",
            "dash_treeview_antd>=0.0.1",
            "fastapi>=0.103.2",
            "uvicorn>=0.30.6"
        ]
    },
    packages=setuptools.find_packages(where = 'src',include=["fusion_tools*"]),
    package_dir={"":"src"},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)


"""Testing image overlay component
"""

import os
import sys
import threading
sys.path.append('./src/')
from fusion_tools.visualization import Visualization
from fusion_tools.components import SlideMap, SlideImageOverlay


def main():

    base_dir = 'C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\'
    local_slide_list = [
        base_dir+'XY01_IU-21-015F.svs',
    ]


    annotations = [
        SlideImageOverlay(
            image_path = 'C:\\Users\\samuelborder\\Downloads\\20250224_095335.jpg'
        )
    ]
    
    vis_session = Visualization(
        local_slides = local_slide_list,
        local_annotations=annotations,
        components = [
            [
                SlideMap()
            ]
        ],
        app_options = {
            'port': 8060
        }
    )

    vis_session.start()


if __name__=='__main__':
    main()

